/*
 * Copyright (c) 1992-1997 Sam Leffler
 * Copyright (c) 1992-1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "tif_conf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "tiffio.h"

#define	CopyField(tag, v, set) \
    if (TIFFGetField(in, tag, &v)) TIFFSetField(out, tag, set)
#define	CopyField2(tag, v1, set1, v2, set2) \
    if (TIFFGetField(in, tag, &v1, &v2)) TIFFSetField(out, tag, set1, set2)
#define	CopyField3(tag, v1, v2, v3) \
    if (TIFFGetField(in, tag, &v1, &v2, &v3)) TIFFSetField(out, tag, v1, v2, v3)

#define PATH_LENGTH 8192

static const char TIFF_SUFFIX[] = ".tif";

static	char fname[PATH_LENGTH];

static	int tiffcp(TIFF*, TIFF*);
static	void newfilename(void);
static	int cpStrips(TIFF*, TIFF*);
static	int cpTiles(TIFF*, TIFF*);

int
main(int argc, char* argv[])
{
	TIFF *in, *out;

	if (argc < 2) {
                fprintf(stderr, "%s\n\n", TIFFGetVersion());
		fprintf(stderr, "usage: tiffsplit input.tif [prefix]\n");
		return 3;
	}

#if defined(TIFF_SLB) && (defined(__atarist__) || defined(__TOS__))
	{
		long ret;
		if ((ret = slb_tiff_open(NULL)) < 0)
		{
			fprintf(stderr, "cannot load tiff.slb: %ld\n", ret);
			return 1;
		}
	}
#endif

	if (argc > 2) {
		strncpy(fname, argv[2], sizeof(fname));
		fname[sizeof(fname) - 1] = '\0';
	}
	in = TIFFOpen(argv[1], "r");
	if (in != NULL) {
		do {
			size_t path_len;
			char *path;
			
			newfilename();

			path_len = strlen(fname) + sizeof(TIFF_SUFFIX);
			path = (char *) _TIFFmalloc(path_len);
			strncpy(path, fname, path_len);
			path[path_len - 1] = '\0';
			strncat(path, TIFF_SUFFIX, path_len - strlen(path) - 1);
			out = TIFFOpen(path, TIFFIsBigEndian(in)?"wb":"wl");
			_TIFFfree(path);

			if (out == NULL)
				return 2;
			if (!tiffcp(in, out))
				return 1;
			TIFFClose(out);
		} while (TIFFReadDirectory(in));
		(void) TIFFClose(in);
	}
	return 0;
}

static void
newfilename(void)
{
	static int first = 1;
	static long lastTurn;
	static long fnum;
	static short defname;
	static char *fpnt;

	if (first) {
		if (fname[0]) {
			fpnt = fname + strlen(fname);
			defname = 0;
		} else {
			fname[0] = 'x';
			fpnt = fname + 1;
			defname = 1;
		}
		first = 0;
	}
#define	MAXFILES	17576
	if (fnum == MAXFILES) {
		if (!defname || fname[0] == 'z') {
			fprintf(stderr, "tiffsplit: too many files.\n");
			exit(1);
		}
		fname[0]++;
		fnum = 0;
	}
	if (fnum % 676 == 0) {
		if (fnum != 0) {
			/*
                         * advance to next letter every 676 pages
			 * condition for 'z'++ will be covered above
                         */
			fpnt[0]++;
		} else {
			/*
                         * set to 'a' if we are on the very first file
                         */
			fpnt[0] = 'a';
		}
		/*
                 * set the value of the last turning point
                 */
		lastTurn = fnum;
	}
	/* 
         * start from 0 every 676 times (provided by lastTurn)
         * this keeps us within a-z boundaries
         */
	fpnt[1] = (char)((fnum - lastTurn) / 26) + 'a';
	/* 
         * cycle last letter every file, from a-z, then repeat
         */
	fpnt[2] = (char)(fnum % 26) + 'a';
	fnum++;
}

static int
tiffcp(TIFF* in, TIFF* out)
{
	uint16 bitspersample, samplesperpixel, compression, shortv, *shortav;
	uint32 w, l;
	float floatv;
	char *stringv;
	uint32 longv;

	CopyField(TIFFTAG_SUBFILETYPE, longv, longv);
	CopyField(TIFFTAG_TILEWIDTH, w, w);
	CopyField(TIFFTAG_TILELENGTH, l, l);
	CopyField(TIFFTAG_IMAGEWIDTH, w, w);
	CopyField(TIFFTAG_IMAGELENGTH, l, l);
	CopyField(TIFFTAG_BITSPERSAMPLE, bitspersample, (tiff_uint_t)bitspersample);
	CopyField(TIFFTAG_SAMPLESPERPIXEL, samplesperpixel, (tiff_uint_t)samplesperpixel);
	CopyField(TIFFTAG_COMPRESSION, compression, (tiff_uint_t)compression);
	if (compression == COMPRESSION_JPEG) {
		uint32 count = 0;
		void *table = NULL;
		if (TIFFGetField(in, TIFFTAG_JPEGTABLES, &count, &table)
		    && count > 0 && table) {
		    TIFFSetField(out, TIFFTAG_JPEGTABLES, count, table);
		}
	}
    CopyField(TIFFTAG_PHOTOMETRIC, shortv, (tiff_uint_t)shortv);
	CopyField(TIFFTAG_PREDICTOR, shortv, (tiff_uint_t)shortv);
	CopyField(TIFFTAG_THRESHHOLDING, shortv, (tiff_uint_t)shortv);
	CopyField(TIFFTAG_FILLORDER, shortv, (tiff_uint_t)shortv);
	CopyField(TIFFTAG_ORIENTATION, shortv, (tiff_uint_t)shortv);
	CopyField(TIFFTAG_MINSAMPLEVALUE, shortv, (tiff_uint_t)shortv);
	CopyField(TIFFTAG_MAXSAMPLEVALUE, shortv, (tiff_uint_t)shortv);
	CopyField(TIFFTAG_XRESOLUTION, floatv, floatv);
	CopyField(TIFFTAG_YRESOLUTION, floatv, floatv);
	CopyField(TIFFTAG_GROUP3OPTIONS, longv, longv);
	CopyField(TIFFTAG_GROUP4OPTIONS, longv, longv);
	CopyField(TIFFTAG_RESOLUTIONUNIT, shortv, (tiff_uint_t)shortv);
	CopyField(TIFFTAG_PLANARCONFIG, shortv, (tiff_uint_t)shortv);
	CopyField(TIFFTAG_ROWSPERSTRIP, longv, longv);
	CopyField(TIFFTAG_XPOSITION, floatv, floatv);
	CopyField(TIFFTAG_YPOSITION, floatv, floatv);
	CopyField(TIFFTAG_IMAGEDEPTH, longv, longv);
	CopyField(TIFFTAG_TILEDEPTH, longv, longv);
	CopyField(TIFFTAG_SAMPLEFORMAT, shortv, (tiff_uint_t)shortv);
	CopyField2(TIFFTAG_EXTRASAMPLES, shortv, (tiff_uint_t)shortv, shortav, shortav);
	{ uint16 *red, *green, *blue;
	  CopyField3(TIFFTAG_COLORMAP, red, green, blue);
	}
	{ uint16 shortv2;
	  CopyField2(TIFFTAG_PAGENUMBER, shortv, (tiff_uint_t)shortv, shortv2, (tiff_uint_t)shortv2);
	}
	CopyField(TIFFTAG_ARTIST, stringv, stringv);
	CopyField(TIFFTAG_IMAGEDESCRIPTION, stringv, stringv);
	CopyField(TIFFTAG_MAKE, stringv, stringv);
	CopyField(TIFFTAG_MODEL, stringv, stringv);
	CopyField(TIFFTAG_SOFTWARE, stringv, stringv);
	CopyField(TIFFTAG_DATETIME, stringv, stringv);
	CopyField(TIFFTAG_HOSTCOMPUTER, stringv, stringv);
	CopyField(TIFFTAG_PAGENAME, stringv, stringv);
	CopyField(TIFFTAG_DOCUMENTNAME, stringv, stringv);
	CopyField(TIFFTAG_BADFAXLINES, longv, longv);
	CopyField(TIFFTAG_CLEANFAXDATA, longv, longv);
	CopyField(TIFFTAG_CONSECUTIVEBADFAXLINES, longv, longv);
	CopyField(TIFFTAG_FAXRECVPARAMS, longv, longv);
	CopyField(TIFFTAG_FAXRECVTIME, longv, longv);
	CopyField(TIFFTAG_FAXSUBADDRESS, stringv, stringv);
	CopyField(TIFFTAG_FAXDCS, stringv, stringv);
	if (TIFFIsTiled(in))
		return (cpTiles(in, out));
	else
		return (cpStrips(in, out));
}

static int
cpStrips(TIFF* in, TIFF* out)
{
	tmsize_t bufsize  = TIFFStripSize(in);
	unsigned char *buf = (unsigned char *)_TIFFmalloc(bufsize);

	if (buf) {
		tstrip_t s, ns = TIFFNumberOfStrips(in);
		uint64 *bytecounts;

		if (!TIFFGetField(in, TIFFTAG_STRIPBYTECOUNTS, &bytecounts)) {
			fprintf(stderr, "tiffsplit: strip byte counts are missing\n");
                        _TIFFfree(buf);
			return (0);
		}
		for (s = 0; s < ns; s++) {
			if (ll_uint64_gt(bytecounts[s], ll_uint64_make(0, bufsize))) {
				buf = (unsigned char *)_TIFFrealloc(buf, ll_uint64_to_size_t(bytecounts[s]));
				if (!buf)
					return (0);
				bufsize = ll_uint64_to_size_t(bytecounts[s]);
			}
			if (TIFFReadRawStrip(in, s, buf, ll_uint64_to_size_t(bytecounts[s])) < 0 ||
			    TIFFWriteRawStrip(out, s, buf, ll_uint64_to_size_t(bytecounts[s])) < 0) {
				_TIFFfree(buf);
				return (0);
			}
		}
		_TIFFfree(buf);
		return (1);
	}
	return (0);
}

static int
cpTiles(TIFF* in, TIFF* out)
{
	tmsize_t bufsize = TIFFTileSize(in);
	unsigned char *buf = (unsigned char *)_TIFFmalloc(bufsize);

	if (buf) {
		ttile_t t, nt = TIFFNumberOfTiles(in);
		uint64 *bytecounts;

		if (!TIFFGetField(in, TIFFTAG_TILEBYTECOUNTS, &bytecounts)) {
			fprintf(stderr, "tiffsplit: tile byte counts are missing\n");
                        _TIFFfree(buf);
			return (0);
		}
		for (t = 0; t < nt; t++) {
			if (ll_uint64_gt(bytecounts[t], ll_uint64_make(0, bufsize))) {
				buf = (unsigned char *)_TIFFrealloc(buf, ll_uint64_to_size_t(bytecounts[t]));
				if (!buf)
					return (0);
				bufsize = ll_uint64_to_size_t(bytecounts[t]);
			}
			if (TIFFReadRawTile(in, t, buf, ll_uint64_to_size_t(bytecounts[t])) < 0 ||
			    TIFFWriteRawTile(out, t, buf, ll_uint64_to_size_t(bytecounts[t])) < 0) {
				_TIFFfree(buf);
				return (0);
			}
		}
		_TIFFfree(buf);
		return (1);
	}
	return (0);
}
